﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.EvolutionTracking {
  [StorableClass]
  [Item("AfterCrossoverOperator", "Performs an action after the crossover operator is applied.")]
  public class BeforeManipulatorOperator<T> : EvolutionTrackingOperator<T>, IManipulatorOperator<T> where T : class, IItem {
    private const string ChildParameterName = "Child";

    public ILookupParameter<T> ChildParameter {
      get { return (ILookupParameter<T>)Parameters[ChildParameterName]; }
    }

    protected BeforeManipulatorOperator(BeforeManipulatorOperator<T> original, Cloner cloner)
      : base(original, cloner) {
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new BeforeManipulatorOperator<T>(this, cloner);
    }

    [StorableConstructor]
    protected BeforeManipulatorOperator(bool deserializing) : base(deserializing) { }

    public BeforeManipulatorOperator() {
      Parameters.Add(new LookupParameter<T>(ChildParameterName));
    }

    public override IOperation Apply() {
      // since mutation always takes place after crossover, the vertex for the current child is already in the tree
      var vChild = GenealogyGraph.GetByContent(ChildParameter.ActualValue);
      var vClone = (IGenealogyGraphNode<T>)vChild.Clone();
      vClone.Rank = vChild.Rank - 0.5;
      GenealogyGraph.AddVertex(vClone);

      var arcs = vChild.InArcs.ToList();
      foreach (var arc in arcs) {
        var p = arc.Source;
        GenealogyGraph.RemoveArc(arc);
        GenealogyGraph.AddArc(new GenealogyGraphArc(p, vClone));
      }
      // by convention, the arc connecting child with parent (non-root parent in case of crossover)
      // holds fragment data when applicable. the following line of code preserves this data 
      // the graph arcs are now connecting parent (rank n) --> clone (rank n+0.5) --> child (rank n+1)
      vClone.InArcs.Last().Data = arcs.Last().Data;
      GenealogyGraph.AddArc(new GenealogyGraphArc(vClone, vChild));

      return base.Apply();
    }
  }
}
